/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.databinding.jaxb;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.wsdl.Binding;
import javax.wsdl.BindingOperation;
import javax.wsdl.Definition;
import javax.wsdl.Input;
import javax.wsdl.Output;
import javax.wsdl.Port;
import javax.wsdl.WSDLException;
import javax.wsdl.factory.WSDLFactory;
import javax.wsdl.xml.WSDLReader;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.Marshaller;
import javax.xml.namespace.QName;

import org.ow2.petals.databinding.jaxb.serialization.ServiceMarshaller;
import org.ow2.petals.databinding.jaxb.serialization.ServiceUnmarshaller;
import org.ow2.petals.databinding.jaxb.service.Message;
import org.ow2.petals.databinding.jaxb.service.Operation;
import org.ow2.petals.databinding.jaxb.service.Service;
import org.ow2.petals.databinding.jaxb.tools.DataBindingConstants;
import org.ow2.petals.databinding.jaxb.tools.Util;

import com.ibm.wsdl.PartImpl;
import com.ibm.wsdl.extensions.soap.SOAPOperationImpl;
import com.ibm.wsdl.extensions.soap12.SOAP12OperationImpl;

/**
 * @author Mohammed EL JAI
 * 
 * This class is a Factory to create an instance of the Service class. 
 * The creation of the Service object is based on WSDL file and generated 
 * JAXB classes.
 */
public class ServiceFactory {
	
	
	/** The single object **/
	private static ServiceFactory wsdl2JbiFactory;
	
	/**
	 * Static method to get a single instance of this class
	 * @return An instance of this class
	 */
	public static ServiceFactory getInstance() {
		if(wsdl2JbiFactory == null) 
			return new ServiceFactory();
		return wsdl2JbiFactory;
	}

	
	/**
	 * Method to create the service object
	 * @param wsdlUrl : The WSDL URL
	 * @param jaxbClassJar : The JAR URL that gathers the generated JAXB classes
	 * @param serviceClassLoader : The service class loader
	 * @return An instance of the Service class
	 * @throws DataBindingException 
	 */
	public Service crateService(
			String wsdlUrl, String jaxbClassJar, 
			ClassLoader serviceClassLoader, Class<?> serviceClass) 
	throws DataBindingException {
			
		Service _service = new Service();
		setWsdlServiceInfo(wsdlUrl, _service);
		setJaxbInfo(jaxbClassJar, _service, serviceClassLoader);
		addMethodSignature(_service, serviceClass);

		return _service;
	}
	
	
	/**
	 * Set WSDL informations to the Service object given in parameter.
	 * @param wsdlUrl : The WSDL URL
	 * @param _service : The service object
	 * @throws DataBindingException 
	 */
	@SuppressWarnings("unchecked")
	private void setWsdlServiceInfo(String wsdlUrl, Service _service) 
	throws DataBindingException {
		
		WSDLFactory wsdlFactory = null;
		Definition definition = null;

		try {
			wsdlFactory = WSDLFactory.newInstance();
		} catch (WSDLException e1) {
			e1.printStackTrace();
		}
		
		// Browse the WSDL Elements
		WSDLReader wsdlReader = wsdlFactory.newWSDLReader();
		try {
			definition = wsdlReader.readWSDL(wsdlUrl);
			Map<QName,javax.wsdl.Service> services = definition.getServices();
			Set<QName> serviceNames = services.keySet();
				
			javax.wsdl.Service service = services.get(serviceNames.iterator().next());
			_service.setServiceName(service.getQName());
			
			Map<String, Port> ports = service.getPorts();
			Set<String> portNames = ports.keySet();
			
			for (String portName : portNames) {
				
				// IF the service has more than one port, choose the soap port
				if (ports.size() > 1 && !portName.endsWith("Soap")) 
					continue;
				
				Port port = (Port) ports.get(portName);
				Binding binding = port.getBinding();
				_service.setInterfaceName(binding.getPortType().getQName());
				List<BindingOperation> bindingOperations = binding.getBindingOperations();
				
				// Browse the binding operations
				for (BindingOperation bindingOperation : bindingOperations) {
					
					// Create service operation
					Operation serviceOperation = new Operation();
					javax.wsdl.Operation operation = bindingOperation.getOperation();
					serviceOperation.setOperationName(operation.getName());
					
					// Create input message 
					Message inputMessage = new Message();
					Input input = operation.getInput();
					javax.wsdl.Message message = input.getMessage();
					
					if (message.getParts().size() > 1)
						throw new DataBindingException("The message element has more than one part, " +
								"only the document/literal wrapped " +
								"WSDL style is supported ");
					
					PartImpl partImpl = (PartImpl) message.getParts().values().iterator().next();
					String partName = partImpl.getName();
					
					inputMessage.setMessageName(message.getPart(partName).getElementName());
					serviceOperation.setInputMessage(inputMessage);
					
					// Create output message
					Message outputMessage = new Message();
					Output output = operation.getOutput();
					message = output.getMessage();
					
					if (message.getParts().size() > 1)
						throw new DataBindingException("The message element has more than one part, " +
								"only the document/literal wrapped " +
								"WSDL style is supported ");
					
					partImpl = (PartImpl) message.getParts().values().iterator().next();
					partName = partImpl.getName();
					
					outputMessage.setMessageName(message.getPart(partName).getElementName());
					serviceOperation.setOutputMessage(outputMessage);
					
					// Get the soap action attribute from the extensible elements
					List extensions = bindingOperation.getExtensibilityElements();
					for (Object extension : extensions) {
						if (extension instanceof SOAPOperationImpl) {
						    SOAPOperationImpl soapOperation = (SOAPOperationImpl) extension;
						    String soapAction = soapOperation.getSoapActionURI();
						    serviceOperation.setSoapActionName(soapAction);
						}
						else if (extensions instanceof SOAP12OperationImpl) {
						    SOAP12OperationImpl soapOperation = (SOAP12OperationImpl) extension;
                            String soapAction = soapOperation.getSoapActionURI();
                            serviceOperation.setSoapActionName(soapAction);
						}
					}
					
					// Add the operation to the service
					_service.getOperations().add(serviceOperation);
				}
			}
		}
					
		catch (Exception e) {
			throw new DataBindingException("WSDL parsing error, can't get service inofrmations" +
					"From the WSDL", e);
		}
		
	}


	/**
	 * Set JAXB informations to the Service object given in parameter. 
	 * @param jaxbClassJar : The JAR URL that gathers the generated JAXB classes 
	 * @param _service : The service object
	 * @param serviceClassLoader : The service class loader 
	 * @throws DataBindingException 
	 */
	private void setJaxbInfo(String jaxbClassJar, Service _service, ClassLoader serviceClassLoader) 
	throws DataBindingException {
		
		ServiceMarshaller serviceMarshaller = new ServiceMarshaller();
		
		ServiceUnmarshaller serviceUnmarshaller = new ServiceUnmarshaller();
		
		try {
			
			ClassLoader jaxbClassLoader = 
				new URLClassLoader(new URL[] {new URL(jaxbClassJar)}, this.getClass().getClassLoader());
			serviceMarshaller.setJaxbClassLoader(jaxbClassLoader);
			serviceUnmarshaller.setServiceClassLoader(serviceClassLoader);			
			
			String packageName = com.sun.tools.xjc.api.XJC.getDefaultPackageName(_service.getServiceName().getNamespaceURI());
			JAXBContext jaxbContext = JAXBContext.newInstance(packageName, jaxbClassLoader);
			serviceMarshaller.setMarshaller(jaxbContext.createMarshaller());
			serviceMarshaller.getMarshaller().setProperty(Marshaller.JAXB_ENCODING, "ISO-8859-1");
			serviceUnmarshaller.setUnmarshaller(jaxbContext.createUnmarshaller());
			
			Class<?> jaxbFactoryClass = jaxbClassLoader.loadClass(packageName + "." + DataBindingConstants.JAXB_FACTORY_CLASS_NAME);
			serviceMarshaller.setJaxbFactoryClass(jaxbFactoryClass);
			
			Object jaxbFactoryInstance = jaxbFactoryClass.newInstance();
			serviceMarshaller.setJaxbFactoryInstance(jaxbFactoryInstance);
			
			List<Operation> serviceOperations = _service.getOperations();
			// Browse the service operations
			for (Operation serviceOperation : serviceOperations) {
				
				// Add JAXB class correspond to input message
				Message inputMessage = serviceOperation.getInputMessage();
				String messageName = Util.upperFirstLetter(
						inputMessage.getMessageName().getLocalPart());
				String jaxbMessageClassName = packageName + "." + messageName;
				
				// Construct the java representation of the input message 
				Class<?> jaxbMessageClass = jaxbClassLoader.loadClass(jaxbMessageClassName);
				inputMessage.setMessageJaxbClass(jaxbMessageClass);
				Field[] fields = jaxbMessageClass.getDeclaredFields();
				for (Field field : fields) 
					inputMessage.getMessageParts().put(field.getName(), field.getType());
				
				// Construct the java representation of the output message
				Message outputMessage = serviceOperation.getOutputMessage();
				messageName = Util.upperFirstLetter(
						outputMessage.getMessageName().getLocalPart());
				jaxbMessageClassName = packageName + "." + messageName;
				
				// construct the java representation of the output message 
				jaxbMessageClass = jaxbClassLoader.loadClass(jaxbMessageClassName);
				outputMessage.setMessageJaxbClass(jaxbMessageClass);
				fields = jaxbMessageClass.getDeclaredFields();
				for (Field field : fields) {
					String filedName = field.getName();
					if (filedName.equals("_return")) 
						filedName = filedName.substring(1);
					outputMessage.getMessageParts().put(filedName, field.getType());
				}
			}
			
			_service.setServiceUnmarshaller(serviceUnmarshaller);
			_service.setServiceMarshaller(serviceMarshaller);

		} catch (Exception e) {
			throw new DataBindingException(" Can't extract the DataBinding inofrmations", e);
		} 
				
	}
	
	
	
	/**
	 * Set method signatures to the service object given in parameter
	 * @param service : The service object.
	 * @param serviceClass : The JAVA interface class of the service
	 */
	private void addMethodSignature(Service service, Class<?> serviceClass) {
        
        List<Operation> serviceOperations = service.getOperations();
           for (Operation serviceOperation : serviceOperations) {
               Method[] methods = serviceClass.getMethods();
               for (Method method : methods) {
                   if (method.getName().equals(serviceOperation.getOperationName())) {
                       serviceOperation.setParameterTypes(method.getParameterTypes());
                       serviceOperation.setGenericParameterTypes(method.getGenericParameterTypes());
                       if (! method.getReturnType().equals(void.class)) {
                           serviceOperation.setReturnType(method.getReturnType());
                           serviceOperation.setGenericReturnType(method.getGenericReturnType());
                       }
                   }
               }
           }
    }
	
}
