/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */


package org.ow2.petals.databinding.jaxb.serialization;

import java.io.StringWriter;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBElement;
import javax.xml.bind.Marshaller;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamResult;

import org.ow2.petals.databinding.jaxb.DataBindingException;
import org.ow2.petals.databinding.jaxb.service.Message;
import org.ow2.petals.databinding.jaxb.service.Operation;
import org.ow2.petals.databinding.jaxb.tools.DataBindingConstants;
import org.ow2.petals.databinding.jaxb.tools.Util;

/**
 * 
 * @author Mohammed EL JAI
 * 
 * This class permits to marshal the request/response message  
 * of the service operations.
 */
public class ServiceMarshaller {

    /**  The class loader of the generated JAXB classes **/
    private ClassLoader jaxbClassLoader;

    /** The JAXB marshaller object **/
    private Marshaller marshaller;

    /** The JAXB factory class to create the messages objects **/
    private Class<?> jaxbFactoryClass;

    /** The JAXB factory instance **/
    private Object jaxbFactoryInstance;



    /**
     * Marshall the method result of the service operation given in parameter
     * @param serviceOperation : The service operation 
     * @param methodResult : The method result 
     * @return a XML message response of the operation given in parameter 
     * @throws DataBindingException
     */
    @SuppressWarnings("unchecked")
    public Source marshallMethodResult(Operation serviceOperation, Object[] methodResult) 
    throws DataBindingException  {

        Source messageContent = null;
        try {

            // Get output message info
            Message outputMessage = serviceOperation.getOutputMessage();
            Map<String,Class<?>> messageParts = outputMessage.getMessageParts();
            String methodName = "create" + Util.upperFirstLetter(outputMessage.getMessageName().getLocalPart());
            Method createJaxbObject = jaxbFactoryClass.getDeclaredMethod(methodName, new Class[]{});
            Object messageJaxbObject = createJaxbObject.invoke(jaxbFactoryInstance, new Object[]{});
            Class<?> messageJaxbClass = outputMessage.getMessageJaxbClass();
            Class<?> returnType = serviceOperation.getReturnType();

            // Browse the method result
            int messageParamIndex = 0;
            for (String messgePartName : messageParts.keySet()) {

                Class<?> messagePartClass = messageParts.get(messgePartName);
                Method setterMethod = null;
                // If the field is a collection type, there is no a setter method
                if (!java.util.Collection.class.isAssignableFrom(messagePartClass)) {
                    String setterMethodName = "set" + Util.upperFirstLetter(messgePartName);
                    setterMethod = messageJaxbClass.getDeclaredMethod(setterMethodName, 
                            new Class[]{messagePartClass});
                }

                if (methodResult[messageParamIndex] == null) {
                    setterMethod.invoke(messageJaxbObject, (Object) null);
                }

                // The method result is a simple type or primitive type
                else if (returnType.isPrimitive() || Util.isSimpleType(returnType)) {

                    setterMethod.invoke(messageJaxbObject, new Object[]{methodResult[messageParamIndex]});

                }

                // The method result is a Calendar type
                else if (returnType.getName().equals(DataBindingConstants.CALENDAR_CLASS_NAME)) {

                    javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                    java.util.Calendar originalBeanDate = (java.util.Calendar) methodResult[messageParamIndex];
                    if (originalBeanDate != null) {
                        jaxbDate = createXMLGregorianCalendarFromDate(originalBeanDate.getTime());
                    }
                    setterMethod.invoke(messageJaxbObject, new Object[]{jaxbDate});

                }

                // The method result is a Date type
                else if (returnType.getName().equals(DataBindingConstants.DATE_CLASS_NAME) ||
                        returnType.getName().equals(DataBindingConstants.SQL_DATE_CLASS_NAME)) {

                    javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                    java.util.Date originalBeanDate = (java.util.Date) methodResult[messageParamIndex];
                    if (originalBeanDate != null) {
                        jaxbDate = createXMLGregorianCalendarFromDate(originalBeanDate);
                    }
                    setterMethod.invoke(messageJaxbObject, new Object[]{jaxbDate});

                }

                // The method result is a collection
                else if (java.util.Collection.class.isAssignableFrom(returnType) || returnType.isArray()) {

                    Object jaxbList = null;
                    List<Object> originalBeanParents = new ArrayList<Object>();
                    List<Object> jaxbBeanParents = new ArrayList<Object> ();
                    Field messageField = Util.getJaxbField(messgePartName, messageJaxbClass.getDeclaredFields());


                    if (java.util.Collection.class.isAssignableFrom(messageField.getType())) {
                        if (setterMethod != null)
                            jaxbList = new ArrayList<Object>();
                        else {
                            String getterMethodName = "get" + Util.upperFirstLetter(messgePartName);
                            Method getterMethod = messageJaxbClass.getDeclaredMethod(getterMethodName);
                            jaxbList = getterMethod.invoke(messageJaxbObject);
                        }
                        if (methodResult[messageParamIndex].getClass().isArray()) {
                            try {
                                createJaxbList(java.util.Arrays.asList((Object[]) methodResult[messageParamIndex]),jaxbList, 
                                        originalBeanParents, jaxbBeanParents);
                            }
                            catch (java.lang.ClassCastException e) {
                                // The array contains a primitive type
                                List<?> listOfWrappedClass = Util.parimitiveArrayToList(methodResult[messageParamIndex]);
                                createJaxbList(listOfWrappedClass,jaxbList, originalBeanParents, jaxbBeanParents);
                            }

                        }
                        else {
                            createJaxbList((Collection<?>) methodResult[messageParamIndex],jaxbList, 
                                    originalBeanParents, jaxbBeanParents);
                        }
                        if (setterMethod != null)
                            setterMethod.invoke(messageJaxbObject,jaxbList);     

                    }

                    else if (messageField.getType().isArray()) {
                        setterMethod.invoke(messageJaxbObject,methodResult[messageParamIndex]);
                    }

                    // The case when the List or the array are wrapped in a JAXB class
                    else {
                        // Get the short class name of the JAXB array class (without package definition)
                        String messagePartShortClassName = messageParts.get(messgePartName).getSimpleName();
                        String createValueMethodName = "create" + messagePartShortClassName;

                        // Get the factory instance to create JAXB array object
                        Class<?> jaxbArrayClass = messageParts.get(messgePartName);
                        Class<?> arrayFactoryClass = jaxbClassLoader.loadClass(
                                jaxbArrayClass.getPackage().getName() + "." + 
                                DataBindingConstants.JAXB_FACTORY_CLASS_NAME);
                        Object arrayFactoryInstance = arrayFactoryClass.newInstance();

                        // Get method from the ObjectFactory class to create the JAXB array instance
                        Method[] arrayFactoryMethods = arrayFactoryClass.getDeclaredMethods();
                        Method createArrayMethod = null;
                        for (Method method : arrayFactoryMethods) {
                            if (method.getName().equals(createValueMethodName)) {
                                createArrayMethod = method;
                            }
                        }
                        jaxbList = createArrayMethod.invoke(arrayFactoryInstance, new Object[]{});
                        if (methodResult[messageParamIndex] != null) {
                            if (!returnType.isArray()) 
                                createJaxbList((Collection<?>) methodResult[messageParamIndex],jaxbList, 
                                        originalBeanParents, jaxbBeanParents);
                            else {
                                try {
                                    createJaxbList(java.util.Arrays.asList(
                                            (Object[]) methodResult[messageParamIndex]),jaxbList, 
                                            originalBeanParents, jaxbBeanParents);
                                }
                                catch(java.lang.ClassCastException e) {
                                    // The array contains a primitive type
                                    List<?> listOfWrappedClass = Util.parimitiveArrayToList(methodResult[messageParamIndex]);
                                    createJaxbList(listOfWrappedClass,jaxbList, originalBeanParents, jaxbBeanParents);
                                }

                            }
                        }
                        setterMethod.invoke(messageJaxbObject,jaxbList);
                    }

                }


                // The method result is a bean
                else {
                    List<Object> originalBeanParents = new ArrayList<Object>();
                    List<Object> jaxbBeanParents = new ArrayList<Object> ();
                    setterMethod.invoke(messageJaxbObject, 
                            new Object[] {createJaxblBean(methodResult[messageParamIndex], 
                                    originalBeanParents, jaxbBeanParents)});
                }
                messageParamIndex ++;
            }

            marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
            StringWriter buffer = new StringWriter();
            Result result = new StreamResult(buffer);

            if (messageJaxbObject.getClass().getAnnotation(XmlRootElement.class) != null)
                marshaller.marshal(messageJaxbObject, result);

            else if(messageJaxbObject.getClass().getAnnotation(XmlType.class) != null) {
                JAXBElement<?> jaxbElement = new JAXBElement(
                        outputMessage.getMessageName(), 
                        messageJaxbObject.getClass(), 
                        messageJaxbObject );
                marshaller.marshal(jaxbElement, result);
            }

            else 
                throw new DataBindingException(" Can't matshall the " + 
                        outputMessage.getMessageName().getLocalPart() + 
                "Missing Jaxb annotation");

            String stringMessageContent =  buffer.toString();
            messageContent = Util.createSourceFromString(stringMessageContent);
        }
        catch(Exception e) {
            throw new DataBindingException(e);
        }



        return messageContent;
    }



    /**
     * Marshall method parameters of the service operation given in parameter
     * @param serviceOperation : The service operation
     * @param methodParams : The method parameters
     * @return : A XML message request to call the operation given in parameter
     * @throws DataBindingException
     */
    @SuppressWarnings("unchecked")
    public Source marshallMethodParam(Operation serviceOperation, Object[] methodParams) throws DataBindingException  {

        Source messageContent = null;

        try {
            // Get output message info
            Message inputMessage = serviceOperation.getInputMessage();
            Map<String,Class<?>> messageParts = inputMessage.getMessageParts();
            String methodName = "create" + Util.upperFirstLetter(inputMessage.getMessageName().getLocalPart());
            Method createJaxbObject = jaxbFactoryClass.getDeclaredMethod(methodName, new Class[]{});
            Object messageJaxbObject = createJaxbObject.invoke(jaxbFactoryInstance, new Object[]{});
            Class<?> messageJaxbClass = inputMessage.getMessageJaxbClass();

            Class<?>[] methodParamTypes = serviceOperation.getParameterTypes();

            // Browse the method parameters
            int messageParamIndex = 0;
            for (String messgePartName : messageParts.keySet()) {

                Class<?> messagePartClass = messageParts.get(messgePartName);
                Method setterMethod = null;
                // If the field is a collection type, there is no a setter method
                if (!java.util.Collection.class.isAssignableFrom(messagePartClass)) {
                    String setterMethodName = "set" + Util.upperFirstLetter(messgePartName);
                    setterMethod = messageJaxbClass.getDeclaredMethod(setterMethodName, 
                            new Class[]{messagePartClass});
                }

                if (methodParams[messageParamIndex] == null) {
                    setterMethod.invoke(messageJaxbObject, (Object) null);
                }

                // the current parameter is a simple type
                else if (methodParamTypes[messageParamIndex].isPrimitive() || Util.isSimpleType(methodParamTypes[messageParamIndex])) {
                    setterMethod.invoke(messageJaxbObject, new Object[]{methodParams[messageParamIndex]});
                }

                // The current parameter is a Calendar type
                else if (methodParamTypes[messageParamIndex].getName().equals(DataBindingConstants.CALENDAR_CLASS_NAME)) {

                    javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                    java.util.Calendar originalBeanDate = (java.util.Calendar) methodParams[messageParamIndex];
                    if (originalBeanDate != null) {
                        jaxbDate = createXMLGregorianCalendarFromDate(originalBeanDate.getTime());
                    }
                    setterMethod.invoke(messageJaxbObject, new Object[]{jaxbDate});

                }

                // The current parameter is a Date type
                else if (methodParamTypes[messageParamIndex].getName().equals(DataBindingConstants.DATE_CLASS_NAME) ||
                        methodParamTypes[messageParamIndex].getName().equals(DataBindingConstants.SQL_DATE_CLASS_NAME)) {

                    javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                    if (methodParams[messageParamIndex] != null) {
                        jaxbDate = createXMLGregorianCalendarFromDate((java.util.Date) methodParams[messageParamIndex]);
                    }
                    setterMethod.invoke(messageJaxbObject, new Object[]{jaxbDate});

                }

                // The current parameter is a list
                else if (java.util.Collection.class.isAssignableFrom(methodParamTypes[messageParamIndex]) || 
                        methodParamTypes[messageParamIndex].isArray()) {

                    Object jaxbList = null;
                    List<Object> originalBeanParents = new ArrayList<Object>();
                    List<Object> jaxbBeanParents = new ArrayList<Object> ();
                    Field messageField = Util.getJaxbField(messgePartName, messageJaxbClass.getDeclaredFields());


                    if (java.util.Collection.class.isAssignableFrom(messageField.getType())) {
                        if (setterMethod != null)
                            jaxbList = new ArrayList<Object>();
                        else {
                            String getterMethodName = "get" + Util.upperFirstLetter(messgePartName);
                            Method getterMethod = messageJaxbClass.getDeclaredMethod(getterMethodName);
                            jaxbList = getterMethod.invoke(messageJaxbObject);
                        }
                        if (methodParams[messageParamIndex].getClass().isArray()) {
                            try {
                                createJaxbList(java.util.Arrays.asList((Object[]) methodParams[messageParamIndex]),jaxbList, 
                                        originalBeanParents, jaxbBeanParents);
                            }
                            catch (java.lang.ClassCastException e) {
                                // The array contains a primitive type
                                List<?> listOfWrappedClass = Util.parimitiveArrayToList(methodParams[messageParamIndex]);
                                createJaxbList(listOfWrappedClass,jaxbList, originalBeanParents, jaxbBeanParents);
                            }

                        }
                        else {
                            createJaxbList((Collection<?>) methodParams[messageParamIndex],jaxbList, 
                                    originalBeanParents, jaxbBeanParents);
                        }
                        if (setterMethod != null)
                            setterMethod.invoke(messageJaxbObject,jaxbList);    

                    }

                    else if (messageField.getType().isArray()) {
                        setterMethod.invoke(messageJaxbObject,methodParams[messageParamIndex]);
                    }

                    // The case when the List or the array are wrapped in a JAXB class
                    else {

                        // Get the short class name of the JAXB array class (without package definition)
                        String messagePartShortClassName = messageParts.get(messgePartName).getSimpleName();
                        String createValueMethodName = "create" + messagePartShortClassName;

                        // Get the factory instance to create JAXB array object
                        Class<?> jaxbArrayClass = messageParts.get(messgePartName);
                        Class<?> arrayFactoryClass = jaxbClassLoader.loadClass(jaxbArrayClass.getPackage().getName() + "." + 
                                DataBindingConstants.JAXB_FACTORY_CLASS_NAME);
                        Object arrayFactoryInstance = arrayFactoryClass.newInstance();

                        // Get method from the ObjectFactory class to create the JAXB array instance
                        Method[] arrayFactoryMethods = arrayFactoryClass.getDeclaredMethods();
                        Method createArrayMethod = null;
                        for (Method method : arrayFactoryMethods) {
                            if (method.getName().equals(createValueMethodName)) {
                                createArrayMethod = method;
                            }
                        }
                        jaxbList = createArrayMethod.invoke(arrayFactoryInstance, new Object[]{});
                        if (methodParams[messageParamIndex] != null) {
                            if (!methodParamTypes[messageParamIndex].isArray())
                                createJaxbList((Collection<?>) methodParams[messageParamIndex],
                                        jaxbList, originalBeanParents, jaxbBeanParents);
                            else {
                                try {
                                    createJaxbList(java.util.Arrays.asList(
                                            (Object[]) methodParams[messageParamIndex]),jaxbList, 
                                            originalBeanParents, jaxbBeanParents);
                                }
                                catch(java.lang.ClassCastException e) {
                                    // The array contains a primitive type
                                    List<?> listOfWrappedClass = Util.parimitiveArrayToList(methodParams[messageParamIndex]);
                                    createJaxbList(listOfWrappedClass,jaxbList, originalBeanParents, jaxbBeanParents);
                                }

                            }
                        }
                        setterMethod.invoke(messageJaxbObject,jaxbList);
                    }

                }
                // The current parameter is a bean
                else {

                    List<Object> originalBeanParents = new ArrayList<Object>();
                    List<Object> jaxbBeanParents = new ArrayList<Object> ();
                    setterMethod.invoke(messageJaxbObject, 
                            new Object[] {createJaxblBean(methodParams[messageParamIndex], 
                                    originalBeanParents, jaxbBeanParents)});

                }
                messageParamIndex ++;
            }

            marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
            StringWriter buffer = new StringWriter();
            Result result = new StreamResult(buffer) ;

            if (messageJaxbObject.getClass().getAnnotation(XmlRootElement.class) != null)
                marshaller.marshal(messageJaxbObject, result);

            else if(messageJaxbObject.getClass().getAnnotation(XmlType.class) != null) {
                JAXBElement<?> jaxbElement = new JAXBElement(
                        inputMessage.getMessageName(), 
                        messageJaxbObject.getClass(), 
                        messageJaxbObject );
                marshaller.marshal(jaxbElement, result);
            }

            String stringMessageContent =  buffer.toString();

            messageContent = Util.createSourceFromString(stringMessageContent);
        }
        catch (Exception e) {
            throw new DataBindingException(e);
        }
        return messageContent;
    }




    /**
     * 
     * @param originalBeanInstance
     * @param originalBeanParents
     * @param jaxbBeanParents
     * @return
     * @throws DataBindingException
     */
    private Object createJaxblBean(Object originalBeanInstance, List<Object> originalBeanParents, 
            List<Object> jaxbBeanParents) 
    throws DataBindingException {

        if (originalBeanInstance == null) return null;

        Object jaxbBeanInstance = null;
        try {

            // Load the JAXB class for the current bean
            String beanClassName = originalBeanInstance.getClass().getSimpleName();
            String beanPackageName = originalBeanInstance.getClass().getPackage().getName();
            Class<?> jaxbBeanClass = null;
            String jaxbBeanClassName = null;
            try {
                jaxbBeanClassName = beanPackageName + "." + beanClassName;
                jaxbBeanClass = jaxbClassLoader.loadClass(jaxbBeanClassName);
            }
            catch(java.lang.NoClassDefFoundError e) {
                // may be that the package name contains uppercase characters
                beanPackageName = beanPackageName.toLowerCase();
                jaxbBeanClassName = beanPackageName + "." + beanClassName;
                jaxbBeanClass = jaxbClassLoader.loadClass(jaxbBeanClassName);
            }


            String simpleClassName = jaxbBeanClass.getSimpleName();
            String methodName = "create" + simpleClassName;

            Class<?> beanFactoryClass = jaxbClassLoader.loadClass(beanPackageName 
                    + "." + DataBindingConstants.JAXB_FACTORY_CLASS_NAME);
            Object beanFactoryInstance = beanFactoryClass.newInstance();
            Method createJaxbObject = beanFactoryClass.getDeclaredMethod(methodName, new Class[]{});
            jaxbBeanInstance = createJaxbObject.invoke(beanFactoryInstance, new Object[]{});

            Field[] originalBeanFields = Util.getFieldFormBeanClassAndSuperClass(originalBeanInstance.getClass());
            for (Field originalBeanField : originalBeanFields) {

                // Initialize the field factory class and instance
                Class<?> fieldFactoryClass = beanFactoryClass;
                Object fieldFactoryInstance = beanFactoryInstance; 

                if (originalBeanField.getName().equals("serialVersionUID")) 
                    continue; 

                Field[] jaxbBeanFields = Util.getFieldFormBeanClassAndSuperClass(jaxbBeanClass);
                Field jaxbBeanField  = Util.getJaxbField(originalBeanField.getName(), jaxbBeanFields);

                if (jaxbBeanField == null)
                    continue;

                Class<?> declaringFieldClass = jaxbBeanField.getDeclaringClass();
                // Check if the current field is an attribute of the super class
                if (!declaringFieldClass.getPackage().getName().equals(beanFactoryClass.getPackage().getName())) {
                    fieldFactoryClass = jaxbClassLoader.loadClass(declaringFieldClass.getPackage().getName() 
                            + "." + DataBindingConstants.JAXB_FACTORY_CLASS_NAME);
                    fieldFactoryInstance = fieldFactoryClass.newInstance();
                }

                String getterMethodName;
                if (!originalBeanField.getType().equals(boolean.class) 
                        && !originalBeanField.getType().equals(java.lang.Boolean.class)) {
                    getterMethodName = "get" + Util.upperFirstLetter(originalBeanField.getName());
                }
                else {
                    getterMethodName = "is" + Util.upperFirstLetter(originalBeanField.getName());
                }


                Method getterMethod = null;
                getterMethod = Util.getMethodFomBeanClassOrSuperClass(getterMethodName, 
                        originalBeanInstance.getClass());

                if (getterMethod == null && getterMethodName.startsWith("is")) {
                    getterMethodName = getterMethodName.replaceFirst("is", "get");
                    getterMethod = Util.getMethodFomBeanClassOrSuperClass(getterMethodName, 
                            originalBeanInstance.getClass());
                }

                if (getterMethod == null) 
                    throw new DataBindingException(" No such method  " + getterMethod);


                String setterMethodName = "set" + Util.upperFirstLetter(jaxbBeanField.getName());
                Method setterMethod = null;
                try {
                 // If the field is a collection type, there is no a setter method
                    if (!java.util.Collection.class.isAssignableFrom(jaxbBeanField.getType()))
                        setterMethod = jaxbBeanClass.getMethod(setterMethodName, 
                                new Class[]{jaxbBeanField.getType()});
                }
                catch(java.lang.NoSuchMethodException e) {
                    setterMethodName = "set" + Util.upperFirstLetter(originalBeanField.getName());
                    setterMethod = jaxbBeanClass.getMethod(setterMethodName, 
                            new Class[]{jaxbBeanField.getType()});
                }

                String createValueMethodName = "create" + declaringFieldClass.getSimpleName() 
                + Util.upperFirstLetter(jaxbBeanField.getName());
                
                // No treatment when the attribute value is null to avoid the serialization of 
                // the null elements in the XML message
                if (getterMethod.invoke(originalBeanInstance, new Object[]{}) == null)
                    continue;

                // The Current field is a primitive type
                if (originalBeanField.getType().isPrimitive() || Util.isSimpleType(originalBeanField.getType())) {

                    if (jaxbBeanField.getType().equals(javax.xml.bind.JAXBElement.class)) {
                        Method createValueMethod = fieldFactoryClass.getDeclaredMethod(createValueMethodName, 
                                new Class[]{originalBeanField.getType()});
                        Object jaxbFieldValue = createValueMethod.invoke(fieldFactoryInstance, 
                                getterMethod.invoke(originalBeanInstance, new Object[]{}));
                        setterMethod.invoke(jaxbBeanInstance, new Object[]{jaxbFieldValue});
                    }
                    else 
                        setterMethod.invoke(jaxbBeanInstance, 
                                getterMethod.invoke(originalBeanInstance, new Object[]{}));

                }


                // The current parameter is a Calendar type
                else if (originalBeanField.getType().getName().
                        equals(DataBindingConstants.CALENDAR_CLASS_NAME)) {

                    javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                    java.util.Calendar originalBeanDate = 
                        (java.util.Calendar) getterMethod.invoke(originalBeanInstance, new Object[]{});
                    Object jaxbFieldValue = null;
                    if (originalBeanDate != null) {
                        jaxbDate = createXMLGregorianCalendarFromDate(originalBeanDate.getTime());
                        Method createValueMethod = fieldFactoryClass.getDeclaredMethod(createValueMethodName, 
                                new Class[]{javax.xml.datatype.XMLGregorianCalendar.class});
                        jaxbFieldValue = createValueMethod.invoke(fieldFactoryInstance, jaxbDate);

                    }
                    setterMethod.invoke(jaxbBeanInstance, new Object[]{jaxbFieldValue});

                }

                // The current parameter is a Date type
                else if (originalBeanField.getType().getName().equals(DataBindingConstants.DATE_CLASS_NAME) ||
                        originalBeanField.getType().getName().equals(DataBindingConstants.SQL_DATE_CLASS_NAME)) {

                    javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                    java.util.Date originalBeanDate = 
                        (java.util.Date) getterMethod.invoke(originalBeanInstance, new Object[]{});
                    if (originalBeanDate != null) {
                        jaxbDate = createXMLGregorianCalendarFromDate(originalBeanDate);
                    }
                    if (jaxbBeanField.getType().equals(javax.xml.bind.JAXBElement.class)) {
                        Method createValueMethod = fieldFactoryClass.getDeclaredMethod(createValueMethodName, 
                                new Class[]{javax.xml.datatype.XMLGregorianCalendar.class});
                        setterMethod.invoke(jaxbBeanInstance, createValueMethod.invoke(fieldFactoryInstance, jaxbDate));

                    }
                    else 
                        setterMethod.invoke(jaxbBeanInstance, new Object[]{jaxbDate});

                }

                // The Current filed is a List
                else if (java.util.Collection.class.isAssignableFrom(originalBeanField.getType()) 
                        || originalBeanField.getType().isArray()) {

                    Object originalBeanList = getterMethod.invoke(originalBeanInstance, new Object[]{});
                    // If the field is null, nothing to do
                    if (originalBeanList == null)
                        continue;
                    
                    if (java.util.Collection.class.isAssignableFrom(jaxbBeanField.getType())) {
                        List<?> jaxbList = null;
                        if (setterMethod != null)
                            jaxbList = new ArrayList<Object>();
                        else {
                            String jaxbGetterMethodName = "get" + Util.upperFirstLetter(jaxbBeanField.getName());
                            Method jaxbgetterMethod = jaxbBeanClass.getMethod(jaxbGetterMethodName);
                            jaxbList = (List<?>) jaxbgetterMethod.invoke(jaxbBeanInstance);
                        }
                        if (originalBeanList.getClass().isArray()) {
                            try {
                                createJaxbList(java.util.Arrays.asList((Object[]) originalBeanList),jaxbList, 
                                        originalBeanParents, jaxbBeanParents);
                            }
                            catch (java.lang.ClassCastException e) {
                                // The array contains a primitive type
                                List<?> listOfWrappedClass = Util.parimitiveArrayToList(originalBeanList);
                                createJaxbList(listOfWrappedClass,jaxbList, originalBeanParents, jaxbBeanParents);
                            }

                        }
                        else {
                            createJaxbList((Collection<?>) originalBeanList,jaxbList, 
                                    originalBeanParents, jaxbBeanParents);
                        }
                        if (setterMethod != null)
                            setterMethod.invoke(jaxbBeanInstance,jaxbList);    
                    }

                    else if (jaxbBeanField.getType().isArray()) {
                        setterMethod.invoke(jaxbBeanInstance,originalBeanList);
                    }

                    // The case when the List or the array are wrapped in a JAXB class
                    else {

                        Method[] jaxbBeanMethods = fieldFactoryClass.getDeclaredMethods();
                        Method createValueMethod = null;
                        for (Method method : jaxbBeanMethods) {
                            if (method.getName().equals(createValueMethodName)) {
                                createValueMethod = method;
                            }
                        }


                        Class<?> jaxbArray = createValueMethod.getParameterTypes()[0];

                        if (jaxbArray.isArray()) {
                            Object jaxbFieldValue = createValueMethod.invoke(fieldFactoryInstance, 
                                    getterMethod.invoke(originalBeanInstance, new Object[]{}));
                            setterMethod.invoke(jaxbBeanInstance, jaxbFieldValue);

                        }
                        else {
                            String jaxbArrayFactoryClassName = jaxbArray.getPackage().getName() + "." + DataBindingConstants.JAXB_FACTORY_CLASS_NAME;
                            Class<?> jaxbArrayFactoryClass = jaxbClassLoader.loadClass(jaxbArrayFactoryClassName);
                            Object jaxbArrayFactoryInstance = jaxbArrayFactoryClass.newInstance();

                            String jaxbArrayClassName = jaxbArray.getSimpleName();
                            String createJaxbArrayMethodName = "create" + jaxbArrayClassName;
                            Method createJaxbArrayMethod = jaxbArrayFactoryClass.getDeclaredMethod(createJaxbArrayMethodName,new Class[]{});
                            Object jaxbArrayObject = createJaxbArrayMethod.invoke(jaxbArrayFactoryInstance, new Object[]{});


                            if (originalBeanList != null) {
                                originalBeanParents.add(originalBeanInstance);
                                jaxbBeanParents.add(jaxbBeanInstance);
                                if (!originalBeanField.getType().isArray())
                                    createJaxbList((Collection<?>)  originalBeanList,jaxbArrayObject, originalBeanParents, jaxbBeanParents);
                                else {
                                    createJaxbList(java.util.Arrays.asList((Object[]) originalBeanList),jaxbArrayObject, originalBeanParents, jaxbBeanParents);
                                }
                            }
                            Object jaxbFieldValue = createValueMethod.invoke(fieldFactoryInstance, jaxbArrayObject);
                            setterMethod.invoke(jaxbBeanInstance, jaxbFieldValue);
                        }
                    }
                }

                else {

                    boolean hasDependence = false;

                    /*for (Object originalBeanParent : originalBeanParents) {
                        // Detect a cycle dependence in the graph object 
                        if (originalBeanParent != null &&  getterMethod.invoke(originalBeanInstance, new Object[]{}) != null &&
                                originalBeanParent.hashCode() == getterMethod.invoke(originalBeanInstance, new Object[]{}).hashCode()) {
                            setterMethod.invoke(jaxbBeanInstance,(Object) null);
                            hasDependence = true;
                        }
                    }*/

                    if (!hasDependence) {
                        originalBeanParents.add(originalBeanInstance);
                        jaxbBeanParents.add(jaxbBeanInstance);
                        Object jaxbBean = 
                            createJaxblBean(getterMethod.invoke(originalBeanInstance, new Object[]{}), 
                                    originalBeanParents, jaxbBeanParents);
                        if (jaxbBean != null) {
                            if (!jaxbBeanField.getType().equals(javax.xml.bind.JAXBElement.class)) {
                                setterMethod.invoke(jaxbBeanInstance, jaxbBean);
                            }
                            // The JAXB field is wrapped in JAXBElement class
                            else {
                                Method createValueMethod = 
                                    fieldFactoryClass.getDeclaredMethod(createValueMethodName, jaxbBean.getClass());
                                Object jaxbFieldValue = 
                                    createValueMethod.invoke(fieldFactoryInstance, jaxbBean);
                                setterMethod.invoke(jaxbBeanInstance, new Object[]{jaxbFieldValue});
                            }
                        }
                        else 
                            setterMethod.invoke(jaxbBeanInstance, new Object[]{null});
                    }
                }

            }
        } catch (Exception e) {
            throw new DataBindingException(e);
        } 

        return jaxbBeanInstance;
    }







    /**
     * 
     * @param originalList
     * @param jaxbArryInsyance
     * @param originalBeanParents
     * @param jaxbBeanParents
     * @throws DataBindingException
     */
    @SuppressWarnings({ "unchecked"})
    private void createJaxbList(Collection<?> originalList, Object jaxbArryInsyance, 
            List<Object> originalBeanParents, List<Object> jaxbBeanParents) throws DataBindingException {

        try {

            Collection<Object> jaxbList ;
            if (!(jaxbArryInsyance instanceof Collection)) {
                jaxbList = 
                    (Collection<Object>) jaxbArryInsyance.getClass().getDeclaredMethods()[0].invoke(jaxbArryInsyance, new Object[0]);
            }
            else
                jaxbList = (Collection<Object>) jaxbArryInsyance;


            for (Object listElement : originalList) {

                if (listElement == null || Util.isSimpleType(listElement.getClass())) {
                    jaxbList.add(listElement);
                }


                // The current element list is a Calendar type
                else if (listElement.getClass().getName().equals(DataBindingConstants.CALENDAR_CLASS_NAME)) {

                    javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                    java.util.Calendar originalDate = (java.util.Calendar) listElement;
                    if (originalDate != null) {
                        jaxbDate = createXMLGregorianCalendarFromDate(originalDate.getTime());

                    }
                    jaxbList.add(jaxbDate);

                }

                // The current element list is a Date type
                else if (listElement.getClass().getName().equals(DataBindingConstants.DATE_CLASS_NAME)) {

                    javax.xml.datatype.XMLGregorianCalendar jaxbDate = null;
                    java.util.Date originalDate = (java.util.Date) listElement;
                    if (originalDate != null) {
                        jaxbDate = createXMLGregorianCalendarFromDate(originalDate);

                    }
                    jaxbList.add(jaxbDate);

                }

                // The current element list is a collection type
                else if(java.util.Collection.class.isAssignableFrom(listElement.getClass())) {

                    Field field = jaxbArryInsyance.getClass().getDeclaredFields()[0];
                    Type type = field.getGenericType();

                    if (type instanceof ParameterizedType) {

                        ParameterizedType pt = (ParameterizedType) type;
                        Type genericType = pt.getActualTypeArguments()[0];

                        Class<?> clazz = null;

                        if (genericType instanceof Class<?>) 
                            clazz = (Class<?>) genericType;

                        else if (genericType instanceof ParameterizedType) 
                            clazz = (Class<?>) ((ParameterizedType) genericType).getRawType();

                        if (clazz != null) {
                            String jaxbArrayFactoryClassName = clazz.getPackage().getName() + "." + DataBindingConstants.JAXB_FACTORY_CLASS_NAME;
                            Class<?> jaxbArrayFactoryClass = jaxbClassLoader.loadClass(jaxbArrayFactoryClassName);
                            Object jaxbArrayFactoryInstance = jaxbArrayFactoryClass.newInstance();

                            String jaxbArrayClassName = clazz.getSimpleName();
                            String createJaxbArrayMethodName = "create" + jaxbArrayClassName;
                            Method createJaxbArrayMethod = jaxbArrayFactoryClass.getDeclaredMethod(createJaxbArrayMethodName,new Class[]{});
                            Object jaxbArrayObject = createJaxbArrayMethod.invoke(jaxbArrayFactoryInstance, new Object[]{});

                            Collection<?> originalList_ = (Collection<?>) listElement;
                            if (originalList_ != null) {

                                createJaxbList(originalList_,jaxbArrayObject, originalBeanParents, jaxbBeanParents);
                            }
                            jaxbList.add(jaxbArrayObject);  


                        }
                        else throw new DataBindingException(" The element list type : " + 
                                listElement.getClass().getName() + "is not supported ");
                    }

                }

                // The current element list is a bean
                else {

                    Object jaxbBeanInstance = createJaxblBean(listElement, new ArrayList(originalBeanParents), new ArrayList(jaxbBeanParents));
                    jaxbList.add(jaxbBeanInstance);
                }
            }
        }
        catch(Exception e) {
            throw new DataBindingException(e);
        }
    }

    /**
     * 
     * @param date
     * @return
     * @throws DatatypeConfigurationException
     */
    private XMLGregorianCalendar createXMLGregorianCalendarFromDate(Date date) throws DatatypeConfigurationException {

        GregorianCalendar gCalendar = new GregorianCalendar();
        gCalendar.setTime(date);
        XMLGregorianCalendar jaxbDate = DatatypeFactory.newInstance().newXMLGregorianCalendar(gCalendar);
        return jaxbDate;

    }



    public ClassLoader getJaxbClassLoader() {
        return jaxbClassLoader;
    }



    public void setJaxbClassLoader(ClassLoader jaxbClassLoader) {
        this.jaxbClassLoader = jaxbClassLoader;
    }



    public Marshaller getMarshaller() {
        return marshaller;
    }



    public void setMarshaller(Marshaller marshaller) {
        this.marshaller = marshaller;
    }



    public Class<?> getJaxbFactoryClass() {
        return jaxbFactoryClass;
    }



    public void setJaxbFactoryClass(Class<?> jaxbFactoryClass) {
        this.jaxbFactoryClass = jaxbFactoryClass;
    }



    public Object getJaxbFactoryInstance() {
        return jaxbFactoryInstance;
    }



    public void setJaxbFactoryInstance(Object jaxbFactoryInstance) {
        this.jaxbFactoryInstance = jaxbFactoryInstance;
    }




}
