/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */


package org.ow2.petals.databinding.jaxb.tools;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;


public class Util {
	
	/**
	 * 
	 * @param word
	 * @return
	 */
	public static String upperFirstLetter(String word) {
		
		String firstLetter = word.substring(0,1);
		String tempWord = word.substring(1);
		firstLetter = firstLetter.toUpperCase();
		word = firstLetter.concat(tempWord);
		return word;
		
	}
	
	/**
	 * 
	 * @param directoryPath
	 */
	public static void createDirectory(String directoryPath) {
		
		File directory = new File(directoryPath);
		if (!directory.exists()) directory.mkdirs();
		
	}
	
	/**
	 *  
	 *  
	 * @param messageContent
	 * @return
	 */
	public static Source createSourceFromString(String messageContent) {

		Source content = null;
		FileOutputStream output;
		try {
			output = new FileOutputStream("temp.xml");
			output.write(messageContent.getBytes());
			output.flush();
			FileInputStream stream = new FileInputStream("temp.xml");
			content = new StreamSource(stream);
			File f = new File("temp.xml");
			f.delete();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}

		return content;

	}
	
	/**
	 * 
	 * @param clazz
	 * @return
	 */
	public static boolean isSimpleType(Class<?> clazz) {
		return clazz.equals(java.lang.String.class) || clazz.equals(java.lang.Integer.class) ||
		clazz.equals(java.lang.Double.class) || clazz.equals(java.lang.Boolean.class) || 
		clazz.equals(java.lang.Long.class) || clazz.equals(java.lang.Character.class) ||
		clazz.equals(java.lang.Short.class) || clazz.equals(java.lang.Byte.class) ||
		clazz.equals(java.lang.Float.class);
	}
	
	
	
	/**
	 * 
	 * @param methodName
	 * @param beanClass
	 * @param methodParamTypes
	 * @return
	 */
	public static Method getMethodFomBeanClassOrSuperClass(String methodName, Class<?> beanClass, Class<?>... methodParamTypes) {

		Class<?> currentBeanClass = beanClass;
		while(currentBeanClass != null) {
			try {
				Method method = currentBeanClass.getDeclaredMethod(methodName, methodParamTypes);
				return method;
			} catch (SecurityException e) {
				return null;
			} catch (NoSuchMethodException e) {
				// loop
			}
			currentBeanClass = currentBeanClass.getSuperclass();
		}
		return null;
	
	}

	
	
	/**
	 * 
	 * @param class1
	 * @return
	 */
	public static  Field[] getFieldFormBeanClassAndSuperClass(
			Class<? extends Object> class1) {

		List<Field> listOfField = new ArrayList<Field>(); 
		Class<?> currentBeanClass = class1;
		while(currentBeanClass != null) {
			listOfField.addAll((java.util.Arrays.asList(currentBeanClass.getDeclaredFields())));
			currentBeanClass = currentBeanClass.getSuperclass();
		}

		Field[] arrayOfField = new Field[listOfField.size()];
		return listOfField.toArray(arrayOfField);
	}
	
	
	/**
	 * 
	 * @param originalBeanFieldName
	 * @param jaxbFields
	 * @return
	 */
	public static Field getJaxbField(String originalBeanFieldName, Field[] jaxbFields) {
		
		for (Field field : jaxbFields) {
		
			XmlElementRef elmentRefAnnotation = field.getAnnotation(XmlElementRef.class);
			XmlElement elementAnnotation = field.getAnnotation(XmlElement.class);
			if (elmentRefAnnotation == null && elementAnnotation == null &&
					originalBeanFieldName.equals(field.getName())) 
				return field;
			else if (elementAnnotation != null && 
			        (elementAnnotation.name().equals(originalBeanFieldName) || 
			                originalBeanFieldName.equals(field.getName())))
			    return field;
			else if (elmentRefAnnotation != null && 
					elmentRefAnnotation.name().equals(originalBeanFieldName)){
				return field;
			}
				
			
		}
		return null;
	}
	
	
	/**
	 * Create a list of wrapped type from the array of 
	 * primitive type given in parameter. If the parameter is a int[] 
	 * the method returns List<Integer>
	 * @param array : array of primitive type
	 * @return List of wrapped Type
	 */
	public static List<?> parimitiveArrayToList(Object array) {
	   
	   Class<?> componentType =  array.getClass().getComponentType();
	   
	   if (componentType.equals(Integer.TYPE)) {
	       int[] arrayOfPrimitiveType = (int[]) array;
	       List<Integer> listOfWrappedType = new ArrayList<Integer>();
	       for (int i=0; i<arrayOfPrimitiveType.length; i++)
	           listOfWrappedType.add(arrayOfPrimitiveType[i]);
	       return listOfWrappedType;
	   }
	   
	   if (componentType.equals(Double.TYPE)) {
           double[] arrayOfPrimitiveType = (double[]) array;
           List<Double> listOfWrappedType = new ArrayList<Double>();
           for (int i=0; i<arrayOfPrimitiveType.length; i++)
               listOfWrappedType.add(arrayOfPrimitiveType[i]);
           return listOfWrappedType;
       }
	   
	   if (componentType.equals(Long.TYPE)) {
           long[] arrayOfPrimitiveType = (long[]) array;
           List<Long> listOfWrappedType = new ArrayList<Long>();
           for (int i=0; i<arrayOfPrimitiveType.length; i++)
               listOfWrappedType.add(arrayOfPrimitiveType[i]);
           return listOfWrappedType;
       }
	   
	   if (componentType.equals(Float.TYPE)) {
           float[] arrayOfPrimitiveType = (float[]) array;
           List<Float> listOfWrappedType = new ArrayList<Float>();
           for (int i=0; i<arrayOfPrimitiveType.length; i++)
               listOfWrappedType.add(arrayOfPrimitiveType[i]);
           return listOfWrappedType;
       }
	   
	   if (componentType.equals(Byte.TYPE)) {
           byte[] arrayOfPrimitiveType = (byte[]) array;
           List<Byte> listOfWrappedType = new ArrayList<Byte>();
           for (int i=0; i<arrayOfPrimitiveType.length; i++)
               listOfWrappedType.add(arrayOfPrimitiveType[i]);
           return listOfWrappedType;
       }
	   
	   if (componentType.equals(Character.TYPE)) {
           char[] arrayOfPrimitiveType = (char[]) array;
           List<Character> listOfWrappedType = new ArrayList<Character>();
           for (int i=0; i<arrayOfPrimitiveType.length; i++)
               listOfWrappedType.add(arrayOfPrimitiveType[i]);
           return listOfWrappedType;
       }
	   
	   if (componentType.equals(Short.TYPE)) {
           short[] arrayOfPrimitiveType = (short[]) array;
           List<Short> listOfWrappedType = new ArrayList<Short>();
           for (int i=0; i<arrayOfPrimitiveType.length; i++)
               listOfWrappedType.add(arrayOfPrimitiveType[i]);
           return listOfWrappedType;
       }
	       
	   return null;
	    
    }
	
	
	/**
	 * Convert a list of a primitive wrapper type in an array 
	 * of primitive type.
	 * example : convert List<Integer> to int[].
	 * @param list : The list of primitive wrapper. 
	 * @param elementArrayType : The primitive type class
	 * @return an array of primitive type
	 */
    public static Object ListToprimitiveArray(List<?> list, Class<?> elementArrayType) {
         
       if (elementArrayType.equals(Integer.TYPE)) {
           int[] arrayOfPrimitiveType = new int[list.size()];
           for (int i=0; i<list.size(); i++) {
               arrayOfPrimitiveType[i] = (Integer)  list.get(i);
           }
           return arrayOfPrimitiveType;
       }
       
       if (elementArrayType.equals(Double.TYPE)) {
           double[] arrayOfPrimitiveType = new double[list.size()];
           for (int i=0; i<list.size(); i++) {
               arrayOfPrimitiveType[i] = (Double)  list.get(i);
           }
           return arrayOfPrimitiveType;
       }
       
       if (elementArrayType.equals(Long.TYPE)) {
           long[] arrayOfPrimitiveType = new long[list.size()];
           for (int i=0; i<list.size(); i++) {
               arrayOfPrimitiveType[i] = (Long)  list.get(i);
           }
           return arrayOfPrimitiveType;
       }
       
       if (elementArrayType.equals(Float.TYPE)) {
           float[] arrayOfPrimitiveType = new float[list.size()];
           for (int i=0; i<list.size(); i++) {
               arrayOfPrimitiveType[i] = (Float)  list.get(i);
           }
           return arrayOfPrimitiveType;
       }
       
       if (elementArrayType.equals(Byte.TYPE)) {
           byte[] arrayOfPrimitiveType = new byte[list.size()];
           for (int i=0; i<list.size(); i++) {
               arrayOfPrimitiveType[i] = (Byte)  list.get(i);
           }
           return arrayOfPrimitiveType;
       }
       
       if (elementArrayType.equals(Character.TYPE)) {
           char[] arrayOfPrimitiveType = new char[list.size()];
           for (int i=0; i<list.size(); i++) {
               arrayOfPrimitiveType[i] = (Character)  list.get(i);
           }
           return arrayOfPrimitiveType;
       }
       
       if (elementArrayType.equals(Short.TYPE)) {
           short[] arrayOfPrimitiveType = new short[list.size()];
           for (int i=0; i<list.size(); i++) {
               arrayOfPrimitiveType[i] = (Short)  list.get(i);
           }
           return arrayOfPrimitiveType;
       }
           
       return null;
        
    }
 	


}
